from types import LambdaType
from . import symbols as s
import operator

symbols = s.SymbolTable()


class Instructions:
    def __init__(self, children):
        if children is None or children == [None]:
            children = []
        self.children = children

    def __len__(self):
        return len(self.children)

    def __iter__(self):
        return iter(self.children)

    def __repr__(self):
        return f'<Instructions {self.children!r}>'

    def eval(self):
        ret = []
        for node in self:
            if isinstance(node, ExitInstruction):
                return node

            result = node.eval()

            if isinstance(result, ExitInstruction):
                return result
            elif result is not None:
                ret.append(result)

        return ret


class BaseExpr:
    def eval(self):
        return NotImplementedError()


class ExitInstruction(BaseExpr):
    def __iter__(self):
        return []

    def eval(self):
        pass


class ReturnInstruction(ExitInstruction):
    def __init__(self, expression: BaseExpr):
        self.expression = expression

    def __repr__(self):
        return f'<ReturnInstruction expression={self.expression!r}>'

    def eval(self):
        return full_eval(self.expression)


class BreakInstruction(ExitInstruction):
    def __repr__(self):
        return f'<BreakInstruction>'

    def eval(self):
        return None


def full_eval(expression: BaseExpr):
    while isinstance(expression, BaseExpr):
        expression = expression.eval()

    return expression


class Primitive(BaseExpr):
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f'<Primitive {self.value!r} ({self.value.__class__.__name__})>'

    def eval(self):
        return self.value


class Identifier(BaseExpr):
    is_func = False

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f'<Identifier {self.name}>'

    def assign(self, value, new=False):
        if self.is_func:
            symbols.set_func(self.name, value)
        else:
            symbols.set_sym(self.name, value, new=new)

    def eval(self):
        if self.is_func:
            return symbols.get_func(self.name)

        return symbols.get_sym(self.name)


class List(BaseExpr):
    def __init__(self, values: Instructions):
        self.values = values

    def __repr__(self):
        return f'<List length={len(self.values)} items={self.values!r}>'

    def eval(self):
        return self.values.eval()


class ListAccess(BaseExpr):
    def __init__(self, list: BaseExpr, index: BaseExpr):
        self.list = list
        self.index = index

    def __repr__(self):
        return f'<List list={self.list!r} index={self.index!r}>'

    def eval(self):
        return self.list.eval()[self.index.eval()]


class ListAssign(BaseExpr):
    def __init__(self, list: Identifier, index: BaseExpr, value: BaseExpr):
        self.list = list
        self.index = index
        self.value = value

    def __repr__(self):
        return f'<List list={self.list!r} index={self.index!r} value={self.value!r}>'

    def eval(self):
        self.list.eval()[self.index.eval()] = self.value.eval()


class ListSlice(BaseExpr):
    def __init__(self, list: BaseExpr, start: BaseExpr, end: BaseExpr):
        self.list = list
        self.start = start
        self.end = end

    def __repr__(self):
        return f'<ListSlice list={self.list!r} start={self.start!r} end={self.end!r}>'

    def eval(self):
        if self.start is not None and self.end is not None:
            return self.list.eval()[self.start.eval():self.end.eval()]
        elif self.start is None and self.end is not None:
            return self.list.eval()[:self.end.eval()]
        elif self.start is not None and self.end is None:
            return self.list.eval()[self.start.eval():]
        else:
            return self.list.eval()


class Range(BaseExpr):
    def __init__(self, start: BaseExpr, end: BaseExpr, inclusive: bool):
        self.start = start
        self.end = end
        self.inclusive = inclusive

    def __repr__(self):
        return f'<Range start={self.start!r} end={self.end!r} inclusive={self.inclusive}>'

    def eval(self):
        if self.inclusive:
            return list(range(self.start.eval(), self.end.eval() + 1))
        else:
            return list(range(self.start.eval(), self.end.eval()))


class Assignment(BaseExpr):
    def __init__(self, identifier: Identifier, value: BaseExpr, new=False):
        self.identifier = identifier
        self.value = value
        self.new = new

    def __repr__(self):
        return f'<Assignment new={self.new} identifier={self.identifier!r} value={self.value!r}>'

    def eval(self):
        if self.identifier.is_func:
            self.identifier.assign(self.value)
        else:
            self.identifier.assign(self.value.eval(), new=self.new)


class CompOp(BaseExpr):
    __ops = {
        '+=': operator.iadd,
        '-=': operator.isub,
        '*=': operator.imul,
        '/=': operator.itruediv,
        '%=': operator.imod,
        '^=': operator.ipow
    }

    def __init__(self, identifier: Identifier, op: str, value: BaseExpr):
        self.identifier = identifier
        self.op = op
        self.value = value

    def __repr__(self):
        return f'<CompOp identifier={self.identifier!r} op={self.op!r} value={self.value!r}>'

    def eval(self):
        left = self.identifier.eval()
        right = self.value.eval()

        try:
            self.identifier.assign(self.__ops[self.op](left, right))
        except TypeError:
            raise TypeError(f'Type error: ({left}: {left.__class__.__name__}) {self.op} ({right}: {right.__class__.__name__}) is not allowed')


class BinOp(BaseExpr):
    __ops = {
        '+': operator.add,
        '-': operator.sub,
        '*': operator.mul,
        '/': operator.truediv,
        '^': operator.pow,
        '%': operator.mod,
        '==': operator.eq,
        '!=': operator.ne,
        '<': operator.lt,
        '>': operator.gt,
        '<=': operator.le,
        '>=': operator.ge,
        '&&': lambda a, b: a.eval() and b.eval(),
        '||': lambda a, b: a.eval() or b.eval()
    }

    def __init__(self, op, left, right):
        self.op = op
        self.left = left
        self.right = right

    def __repr__(self):
        return f'<BinOp op={self.op!r} left={self.left!r} right={self.right!r}>'

    def eval(self):
        left = None
        right = None

        try:
            op = self.__ops[self.op]

            if isinstance(op, LambdaType):
                return op(self.left, self.right)

            left = self.left.eval()
            right = self.right.eval()

            return op(left, right)
        except TypeError:
            # TODO: custom error
            raise TypeError(f'Type error: ({left}: {left.__class__.__name__}) {self.op} ({right}: {right.__class__.__name__}) is not allowed')


class UnaryOp(BaseExpr):
    __ops = {
        '-': operator.neg,
        '+': operator.abs,
        '!': operator.not_
    }

    def __init__(self, op, value: BaseExpr):
        self.op = op
        self.value = value

    def __repr__(self):
        return f'<UnaryOp op={self.op!r} value={self.value!r}>'

    def eval(self):
        return self.__ops[self.op](self.value.eval())


class If(BaseExpr):
    def __init__(self, condition: BaseExpr, true_branch: Instructions, false_branch: Instructions = None):
        self.condition = condition
        self.true_branch = true_branch
        self.false_branch = false_branch

    def __repr__(self):
        return f'<If condition={self.condition!r} true_branch={self.true_branch!r} false_branch={self.false_branch!r}>'

    def eval(self):
        if self.condition.eval():
            return self.true_branch.eval()
        elif self.false_branch is not None:
            return self.false_branch.eval()


class For(BaseExpr):
    def __init__(self, var: Identifier, sequence: BaseExpr, body: Instructions):
        self.var = var
        self.sequence = sequence
        self.body = body

    def __repr__(self):
        return f'<For var={self.var!r} sequence={self.sequence!r} body={self.body!r}>'

    def eval(self):
        for value in self.sequence.eval():
            self.var.assign(value)
            val = self.body.eval()
            if isinstance(val, ReturnInstruction):
                return val
            elif isinstance(val, BreakInstruction):
                break


class While(BaseExpr):
    def __init__(self, condition: BaseExpr, body: Instructions):
        self.condition = condition
        self.body = body

    def __repr__(self):
        return f'<While condition={self.condition!r} body={self.body!r}>'

    def eval(self):
        while self.condition.eval():
            val = self.body.eval()
            if isinstance(val, ReturnInstruction):
                return val
            elif isinstance(val, BreakInstruction):
                break


class Function(BaseExpr):
    def __init__(self, params: Instructions, body: Instructions):
        self.params = params
        self.body = body

    def __repr__(self):
        return f'<Function params={self.params!r} body={self.body!r}>'

    def eval(self, args):
        symbols.set_local(True)

        for key, value in args.items():
            symbols.set_sym(key, value, new=True)

        try:
            result = self.body.eval()

            if isinstance(result, ReturnInstruction):
                return result.eval()
        finally:
            symbols.set_local(False)

        return None


class FunctionCall(BaseExpr):
    def __init__(self, name: Identifier, params: Instructions):
        self.name = name
        self.params = params

    def __repr__(self):
        return f'<Function call name={self.name!r} params={self.params!r}>'

    def __builtin_func(self):
        func = self.name.eval()
        args = []

        for param in self.params:
            args.append(full_eval(param))

        return func.eval(args)

    def __user_func(self):
        func = self.name.eval()
        args = {}

        length1 = len(func.params)
        length2 = len(self.params)

        if length1 != length2:
            raise NameError(f'Invalid amount of args for "{self.name.name}". Got {length2}, expected {length1}')

        for param, value in zip(func.params, self.params):
            args[param.name] = full_eval(value)

        return func.eval(args)

    def eval(self):
        if isinstance(self.name.eval(), BuiltinFunc):
            return self.__builtin_func()

        return self.__user_func()


class BuiltinFunc(BaseExpr):
    def __init__(self, func):
        self.func = func

    def __repr__(self):
        return f'<Builtin func {self.func!r}>'

    def eval(self, args):
        return self.func(*args)


import lark
from . import nodes


class Transformer(lark.Transformer):
    def start(self, args):
        return nodes.Instructions(args)

    def codeblock(self, args):
        return nodes.Instructions(args)

    def num(self, args):
        return nodes.Primitive(int(args[0]))

    def string(self, args):
        s = args[0][1:-1]
        s = bytes(s, "utf-8").decode("unicode_escape")
        return nodes.Primitive(s)

    def var(self, args):
        return nodes.Identifier(args[0])

    def true(self, args):
        return nodes.Primitive(True)

    def false(self, args):
        return nodes.Primitive(False)

    def list(self, args):
        return nodes.List(nodes.Instructions(args))

    def list_access(self, args):
        return nodes.ListAccess(args[0], args[1])

    def list_slice(self, args):
        return nodes.ListSlice(args[0], args[1], args[2])

    def list_assign(self, args):
        return nodes.ListAssign(nodes.Identifier(args[0].value), args[1], args[3])

    def range_incl(self, args):
        return nodes.Range(args[0], args[1], True)

    def range_excl(self, args):
        return nodes.Range(args[0], args[1], False)

    def neg(self, args):
        return nodes.UnaryOp('-', args[0])

    def abs(self, args):
        return nodes.UnaryOp('+', args[0])

    def not_(self, args):
        return nodes.UnaryOp('!', args[0])

    def add(self, args):
        return nodes.BinOp('+', args[0], args[1])

    def sub(self, args):
        return nodes.BinOp('-', args[0], args[1])

    def mul(self, args):
        return nodes.BinOp('*', args[0], args[1])

    def div(self, args):
        return nodes.BinOp('/', args[0], args[1])

    def mod(self, args):
        return nodes.BinOp('%', args[0], args[1])

    def pow(self, args):
        return nodes.BinOp('^', args[0], args[1])

    def eq(self, args):
        return nodes.BinOp('==', args[0], args[1])

    def neq(self, args):
        return nodes.BinOp('!=', args[0], args[1])

    def lt(self, args):
        return nodes.BinOp('<', args[0], args[1])

    def gt(self, args):
        return nodes.BinOp('>', args[0], args[1])

    def lteq(self, args):
        return nodes.BinOp('<=', args[0], args[1])

    def gteq(self, args):
        return nodes.BinOp('>=', args[0], args[1])

    def and_(self, args):
        return nodes.BinOp('&&', args[0], args[1])

    def or_(self, args):
        return nodes.BinOp('||', args[0], args[1])

    def assign_var(self, args):
        # Variable assignments don't have to have a value, like in `this x`.
        # If they don't, the value of the assignment is None.
        if args[1] is None:
            args[1] = nodes.Primitive(None)
        return nodes.Assignment(nodes.Identifier(args[0]), args[1], new=True)

    def assign_value(self, args):
        if args[1] == '=':
            return nodes.Assignment(nodes.Identifier(args[0]), args[2])
        else:
            return nodes.CompOp(nodes.Identifier(args[0]), args[1].value, args[2])

    def if_(self, args):
        return nodes.If(args[0], args[1])

    def if_else(self, args):
        return nodes.If(args[0], args[1], args[2])

    def for_(self, args):
        return nodes.For(nodes.Identifier(args[0].value), args[1], args[2])

    def while_(self, args):
        return nodes.While(args[0], args[1])

    def break_(self, args):
        return nodes.BreakInstruction()

    def return_(self, args):
        return nodes.ReturnInstruction(args[0])

    def fun_def(self, args):
        id = nodes.Identifier(args[0].value)
        id.is_func = True
        return nodes.Assignment(id, nodes.Function(args[1], args[2]), new=True)

    def fun_args(self, args):
        # args[0] can be None, meaning there are no arguments
        if args[0] is None:
            temp = None
        else:
            temp = [nodes.Identifier(arg.value) for arg in args]
        return nodes.Instructions(temp)

    def fun_call(self, args):
        id = nodes.Identifier(args[0].value)
        id.is_func = True
        return nodes.FunctionCall(id, nodes.Instructions(args[1:]))


from lark import Lark, UnexpectedInput
from .transformer import Transformer
import pickle, os, tinted
from . import environment


def parse(code):
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'syntax.lark'), 'r') as f:
        syntax = f.read()
    parser = Lark(syntax)
    try:
        return parser.parse(code)
    except UnexpectedInput as e:
        print(tinted.tint(f'[red][bold]Syntax error[/][/] at [blue][bold]line[/][/] {e.line}, [blue][bold]column[/][/] {e.column}\n\n{e.get_context(code)}'))
        exit(1)


def transform(tree):
    transformer = Transformer()
    return transformer.transform(tree)


def execute(transformed):
    environment.define_builtins()
    return transformed.eval()


def talon(inputfile: str, compile=False, outputfile=None):
    if inputfile.endswith('.tal'):
        with open(inputfile, 'r') as input:
            code = input.read()
            temp = transform(parse(code))
    elif inputfile.endswith('.talc'):
        with open(inputfile, 'rb') as f:
            temp = pickle.load(f)
    else:
        print(tinted.tint('[red][bold]Error[/][/]: Invalid file extension'))
        exit(1)
    try:
        if compile:
            if outputfile is None:
                outputfile = inputfile + 'c'
            else:
                if not outputfile.endswith('.talc'):
                    outputfile += '.talc'
            with open(outputfile, 'wb') as f:
                pickle.dump(temp, f)
        else:
            execute(temp)

    except Exception as e:
        print(tinted.tint(f'[red][bold]{str(e.__class__.__name__)}[/][/]: {str(e)}'))
        exit(1)


def main():
    import sys
    if len(sys.argv) == 2:
        talon(sys.argv[1])
    elif len(sys.argv) == 3 and sys.argv[1] == '-c':
        talon(sys.argv[2], compile=True)
    elif len(sys.argv) == 5 and sys.argv[1] == '-c' and sys.argv[3] == '-o':
        talon(sys.argv[2], compile=True, outputfile=sys.argv[4])
    else:
        print('Usage: talon [-c] <input.tal[c]> [-o <output.talc>]')
        exit(1)


if __name__ == '__main__':
    main()


class VMError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class SymbolNotFound(VMError):
    pass


class SymbolExists(VMError):
    pass


class FunctionNotFound(VMError):
    pass


class FunctionExists(VMError):
    pass


class WrongParamCount(VMError):
    pass


import math, random, re, tinted
from . import nodes, talon


def format(string, list):
    def repl(match):
        if match[0][0:2] == '%%':
            return match[0][1:]
        return str(list[int(match[1])])

    string = re.sub(r'%?%(-?\d+)', repl, string)
    return string


def colored(string):
    return tinted.tint(string)


def printf(string, list):
    print(format(string, list))


def printc(string):
    print(colored(string))


def import_(name):
    with open(name, 'r') as f:
        talon.transform(talon.parse(f.read())).eval()


def define_builtins():
    table = nodes.symbols
    b = nodes.BuiltinFunc

    table.set_func('print', b(print))
    table.set_func('printf', b(printf))
    table.set_func('printc', b(printc))
    table.set_func('getstr', b(input))
    table.set_func('format', b(format))
    table.set_func('colored', b(colored))
    table.set_func('import', b(import_))

    # Math
    table.set_func('int', b(int))
    table.set_func('float', b(float))
    table.set_func('round', b(round))
    table.set_func('log', b(math.log))
    table.set_func('sqrt', b(math.sqrt))
    table.set_func('sin', b(math.sin))
    table.set_func('cos', b(math.cos))
    table.set_func('tan', b(math.tan))
    table.set_func('asin', b(math.asin))
    table.set_func('acos', b(math.acos))
    table.set_func('atan', b(math.atan))
    table.set_func('atan2', b(math.atan2))
    table.set_func('random', b(random.random))
    table.set_func('randomint', b(random.randint))

    # Strings
    table.set_func('str', b(str))
    table.set_func('len', b(len))
    table.set_func('ord', b(ord))
    table.set_func('chr', b(chr))
    table.set_func('lower', b(str.lower))
    table.set_func('upper', b(str.upper))
    table.set_func('startswith', b(str.startswith))
    table.set_func('endswith', b(str.endswith))
    table.set_func('replace', b(str.replace))
    table.set_func('split', b(str.split))

    # Lists
    table.set_func('list', b(list))
    table.set_func('append', b(lambda l, e: l.append(e)))
    table.set_func('pop', b(lambda l: l.pop()))
    table.set_func('push', b(lambda l, e: l.push(e)))
    table.set_func('remove', b(lambda l, e: l.remove(e)))
    table.set_func('reverse', b(lambda l: l.reverse()))
    table.set_func('sort', b(lambda l: l.sort()))



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="talonlang",
    version="1.0.0",
    description="The Talon programming language",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/MystPi/talon",
    author="MystPi",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3"
    ],
    packages=["talon"],
    package_data={'': ['*.lark']},
    install_requires=["lark", "tinted"],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "tal=talon.talon:main",
        ]
    },
)


from .errors import *


class SymbolTable:
    __func  = 'functions'
    __sym   = 'symbols'
    __local = 'locals'

    __table = {__func: {}, __sym: {}, __local: []}

    def __is_local(self):
        return len(self.__table[self.__local]) > 0

    def table(self):
        return self.__table

    def get_local_table(self):
        return self.__table[self.__local][-1]

    def set_local(self, flag):
        if flag:
            self.__table[self.__local].append({})
        else:
            self.__table[self.__local].pop()

    def get_sym(self, sym):
        if self.__is_local():
            for table in reversed(self.__table[self.__local]):
                if sym in table:
                    return table[sym]

        if sym in self.__table[self.__sym]:
            return self.__table[self.__sym][sym]

        raise SymbolNotFound(f'Symbol "{sym}" not found')

    def set_sym(self, sym, val, new=False):
        if new and self.__is_local() and sym in self.get_local_table():
            raise SymbolExists(f'Cannot recreate variable "{sym}"')
        elif new and sym in self.__table[self.__sym] and not self.__is_local():
            raise SymbolExists(f'Cannot recreate variable "{sym}"')
        elif new:
            if self.__is_local():
                self.get_local_table()[sym] = val
            else:
                self.__table[self.__sym][sym] = val
        else:
            if self.__is_local():
                for table in reversed(self.__table[self.__local]):
                    if sym in table:
                        table[sym] = val
                        return
                if sym in self.__table[self.__sym]:
                    self.__table[self.__sym][sym] = val
                    return
                raise SymbolNotFound(f'Symbol "{sym}" not found')
            else:
                if sym in self.__table[self.__sym]:
                    self.__table[self.__sym][sym] = val
                else:
                    raise SymbolNotFound(f'Symbol "{sym}" not found')

    def get_func(self, func):
        if func in self.__table[self.__func]:
            return self.__table[self.__func][func]

        raise FunctionNotFound(f'Function "{func}" not found')

    def set_func(self, func, val):
        if func in self.__table[self.__func]:
            raise FunctionExists(f'Function "{func}" cannot be redefined')
        
        self.__table[self.__func][func] = val